@@ -13,7 +13,6 @@
 (*                                                                        *)
 (**************************************************************************)
 
-open Compenv
 open Parsetree
 module String = Misc.Stdlib.String
 
@@ -31,6 +30,7 @@ let bytecode_only = ref false
 let raw_dependencies = ref false
 let sort_files = ref false
 let all_dependencies = ref false
+let nocwd = ref false
 let one_line = ref false
 let files =
   ref ([] : (string * file_kind * String.Set.t * string list) list)
@@ -310,7 +310,12 @@ let read_parse_and_extract parse_function extract_function def ast_kind
       let bound_vars =
         List.fold_left
           (fun bv modname ->
-            Depend.open_module bv (Longident.parse modname))
+             let lid =
+               let lexbuf = Lexing.from_string modname in
+               Location.init lexbuf
+                 (Printf.sprintf "command line argument: -open %S" modname);
+               Parse.simple_module_path lexbuf in
+             Depend.open_module bv lid)
           !module_map ((* PR#7248 *) List.rev !Clflags.open_modules)
       in
       let r = extract_function bound_vars ast in
@@ -401,10 +406,12 @@ let mli_file_dependencies source_file =
 let process_file_as process_fun def source_file =
   Compenv.readenv ppf (Before_compile source_file);
   load_path := [];
+  let cwd = if !nocwd then [] else [Filename.current_dir_name] in
   List.iter add_to_load_path (
       (!Compenv.last_include_dirs @
        !Clflags.include_dirs @
-       !Compenv.first_include_dirs
+       !Compenv.first_include_dirs @
+       cwd
       ));
   Location.input_name := source_file;
   try
@@ -554,6 +561,18 @@ let parse_map fname =
   module_map := String.Map.add modname mm !module_map
 ;;
 
+(* Dependency processing *)
+
+type dep_arg =
+  | Map of Misc.filepath (* -map option *)
+  | Src of Misc.filepath * file_kind option (* -impl, -intf or anon arg *)
+
+let process_dep_arg = function
+  | Map file -> parse_map file
+  | Src (file, None) -> file_dependencies file
+  | Src (file, (Some file_kind)) -> file_dependencies_as file_kind file
+
+let process_dep_args dep_args = List.iter process_dep_arg dep_args
 
 (* Entry point *)
 
@@ -567,77 +586,92 @@ let print_version_num () =
   exit 0;
 ;;
 
-let main () =
+
+let run_main argv =
+  let dep_args_rev : dep_arg list ref = ref [] in
+  let add_dep_arg f s = dep_args_rev := (f s) :: !dep_args_rev in
   Clflags.classic := false;
-  add_to_list first_include_dirs Filename.current_dir_name;
-  Compenv.readenv ppf Before_args;
-  Clflags.reset_arguments (); (* reset arguments from ocamlc/ocamlopt *)
-  Clflags.add_arguments __LOC__ [
-     "-absname", Arg.Set Clflags.absname,
+  try
+    Compenv.readenv ppf Before_args;
+    Clflags.reset_arguments (); (* reset arguments from ocamlc/ocamlopt *)
+    Clflags.add_arguments __LOC__ [
+      "-absname", Arg.Set Clflags.absname,
         " Show absolute filenames in error messages";
-     "-all", Arg.Set all_dependencies,
+      "-all", Arg.Set all_dependencies,
         " Generate dependencies on all files";
-     "-allow-approx", Arg.Set allow_approximation,
+      "-allow-approx", Arg.Set allow_approximation,
         " Fallback to a lexer-based approximation on unparsable files";
-     "-as-map", Arg.Set Clflags.transparent_modules,
-      " Omit delayed dependencies for module aliases (-no-alias-deps -w -49)";
-      (* "compiler uses -no-alias-deps, and no module is coerced"; *)
-     "-debug-map", Arg.Set debug,
+      "-as-map", Arg.Set Clflags.transparent_modules,
+        " Omit delayed dependencies for module aliases (-no-alias-deps -w -49)";
+        (* "compiler uses -no-alias-deps, and no module is coerced"; *)
+      "-debug-map", Arg.Set debug,
         " Dump the delayed dependency map for each map file";
-     "-I", Arg.String (add_to_list Clflags.include_dirs),
+      "-I", Arg.String (add_to_list Clflags.include_dirs),
         "<dir>  Add <dir> to the list of include directories";
-     "-impl", Arg.String (file_dependencies_as ML),
+      "-nocwd", Arg.Set nocwd,
+        " Do not add current working directory to \
+          the list of include directories";
+      "-impl", Arg.String (add_dep_arg (fun f -> Src (f, Some ML))),
         "<f>  Process <f> as a .ml file";
-     "-intf", Arg.String (file_dependencies_as MLI),
+      "-intf", Arg.String (add_dep_arg (fun f -> Src (f, Some MLI))),
         "<f>  Process <f> as a .mli file";
-     "-map", Arg.String parse_map,
+      "-map", Arg.String (add_dep_arg (fun f -> Map f)),
         "<f>  Read <f> and propagate delayed dependencies to following files";
-     "-ml-synonym", Arg.String(add_to_synonym_list ml_synonyms),
+      "-ml-synonym", Arg.String(add_to_synonym_list ml_synonyms),
         "<e>  Consider <e> as a synonym of the .ml extension";
-     "-mli-synonym", Arg.String(add_to_synonym_list mli_synonyms),
+      "-mli-synonym", Arg.String(add_to_synonym_list mli_synonyms),
         "<e>  Consider <e> as a synonym of the .mli extension";
-     "-modules", Arg.Set raw_dependencies,
+      "-modules", Arg.Set raw_dependencies,
         " Print module dependencies in raw form (not suitable for make)";
-     "-native", Arg.Set native_only,
+      "-native", Arg.Set native_only,
         " Generate dependencies for native-code only (no .cmo files)";
-     "-bytecode", Arg.Set bytecode_only,
+      "-bytecode", Arg.Set bytecode_only,
         " Generate dependencies for bytecode-code only (no .cmx files)";
-     "-one-line", Arg.Set one_line,
+      "-one-line", Arg.Set one_line,
         " Output one line per file, regardless of the length";
-     "-open", Arg.String (add_to_list Clflags.open_modules),
+      "-open", Arg.String (add_to_list Clflags.open_modules),
         "<module>  Opens the module <module> before typing";
-     "-plugin", Arg.String(fun _p -> Clflags.plugin := true),
-         "<plugin>  (no longer supported)";
-     "-pp", Arg.String(fun s -> Clflags.preprocessor := Some s),
-         "<cmd>  Pipe sources through preprocessor <cmd>";
-     "-ppx", Arg.String (add_to_list first_ppx),
-         "<cmd>  Pipe abstract syntax trees through preprocessor <cmd>";
-     "-shared", Arg.Set shared,
-         " Generate dependencies for native plugin files (.cmxs targets)";
-     "-slash", Arg.Set Clflags.force_slash,
-         " (Windows) Use forward slash / instead of backslash \\ in file paths";
-     "-sort", Arg.Set sort_files,
+      "-plugin", Arg.String(fun _p -> Clflags.plugin := true),
+        "<plugin>  (no longer supported)";
+      "-pp", Arg.String(fun s -> Clflags.preprocessor := Some s),
+        "<cmd>  Pipe sources through preprocessor <cmd>";
+      "-ppx", Arg.String (add_to_list Compenv.first_ppx),
+        "<cmd>  Pipe abstract syntax trees through preprocessor <cmd>";
+      "-shared", Arg.Set shared,
+        " Generate dependencies for native plugin files (.cmxs targets)";
+      "-slash", Arg.Set Clflags.force_slash,
+        " (Windows) Use forward slash / instead of backslash \\ in file paths";
+      "-sort", Arg.Set sort_files,
         " Sort files according to their dependencies";
-     "-version", Arg.Unit print_version,
-         " Print version and exit";
-     "-vnum", Arg.Unit print_version_num,
-         " Print version number and exit";
-     "-args", Arg.Expand Arg.read_arg,
-         "<file> Read additional newline separated command line arguments \n\
-         \      from <file>";
-     "-args0", Arg.Expand Arg.read_arg0,
-         "<file> Read additional NUL separated command line arguments from \n\
-         \      <file>"
-  ];
-  let usage =
-    Printf.sprintf "Usage: %s [options] <source files>\nOptions are:"
-                   (Filename.basename Sys.argv.(0))
-  in
-  Clflags.parse_arguments file_dependencies usage;
-  Compenv.readenv ppf Before_link;
-  if !sort_files then sort_files_by_dependencies !files
-  else List.iter print_file_dependencies (List.sort compare !files);
-  exit (if Error_occurred.get () then 2 else 0)
+      "-version", Arg.Unit print_version,
+        " Print version and exit";
+      "-vnum", Arg.Unit print_version_num,
+        " Print version number and exit";
+      "-args", Arg.Expand Arg.read_arg,
+        "<file> Read additional newline separated command line arguments \n\
+        \      from <file>";
+      "-args0", Arg.Expand Arg.read_arg0,
+        "<file> Read additional NUL separated command line arguments from \n\
+        \      <file>"
+    ];
+    let program = Filename.basename Sys.argv.(0) in
+    Compenv.parse_arguments (ref argv)
+      (add_dep_arg (fun f -> Src (f, None))) program;
+    process_dep_args (List.rev !dep_args_rev);
+    Compenv.readenv ppf Before_link;
+    if !sort_files then sort_files_by_dependencies !files
+    else List.iter print_file_dependencies (List.sort compare !files);
+    (if Error_occurred.get () then 2 else 0)
+  with
+  | Compenv.Exit_with_status n ->
+      n
+  | exn ->
+      Location.report_exception ppf exn;
+      2
+
+
+let main () =
+  exit (run_main Sys.argv)
 
 let main_from_option () =
   if Sys.argv.(1) <> "-depend" then begin
@@ -645,7 +679,8 @@ let main_from_option () =
       "Fatal error: argument -depend must be used as first argument.\n%!";
     exit 2;
   end;
-  incr Arg.current;
-  Sys.argv.(0) <- Sys.argv.(0) ^ " -depend";
-  Sys.argv.(!Arg.current) <- Sys.argv.(0);
-  main ()
+  let args =
+    Array.concat [ [| Sys.argv.(0) ^ " -depend" |];
+                   Array.sub Sys.argv 2 (Array.length Sys.argv - 2) ] in
+  Sys.argv.(0) <- args.(0);
+  exit (run_main args)
